import argparse
import functools
import pathlib

import qpformat

from .._version import version

from . import definitions
from . import config
from .profile import get_profile_path

#: DryMass analysis output suffix (appended to data path)
OUTPUT_SUFFIX = "_dm"

META_MAPPER = {"medium index": ("medium index", 1),
               "pixel size um": ("pixel size", 1e6),
               "wavelength nm": ("wavelength", 1e9),
               }


@functools.lru_cache(maxsize=32)
def input_setting(path, section, key):
    """Ask the user for a configuration key"""
    cfg = config.ConfigFile(path)
    sec = cfg[section]
    description = definitions.config[section][key][2]
    if key not in sec or sec[key] is None:
        val = input("Please enter '{}' ({}): ".format(key, description))
        cfg.set_value("meta", key, val)


def main(path=None, req_meta=[], description="DryMass analysis.",
         profile=None, recursive=False):
    """Main user dialog with optional "meta" kwargs required

    Parameters
    ----------
    path: str, pathlib.Path, or None
        Path to the measurement data. If set to `None`, the
        command-line will be parsed.
    req_meta: list of str
        Keyword arguments of the [meta] section in drymass.cfg that
        are required by the current task.
    description: str
        Description of the current task. The description is
        displayed when the user executes a console_script
        entry-point with the `--help` argument.
    profile: str, pathlib.Path, or None
        A path to a 'drymass.cfg' file or a name of a profile in
        the local library (see :mod:`drymass.cli.profile`). If set
        to `None`, the default profile is used and the user is asked
        for missing values.
    recursive: bool
        Perform recursive search in `path`. If `path` is None, then
        `recursive` must be False. Instead, the `recursive` argument
        should be set via the command line.

    Returns
    -------
    path_in: pathlib.Path or list of pathlib.Path
        The measurement path. If a recursive search is performed
        (see `recursive` parameter above), then a list of the
        measurement paths is returned.
    path_out: pathlib.Path or None
        The output path, i.e. the path with `_dm` appended. If a
        recursive search is performed, `path_out` is set to None.
    """
    # get directories
    if path is None:
        if recursive:
            msg = "'recursive' must not be set when 'path' is 'None'!"
            raise ValueError(msg)
        path_in, profile, recursive = parse(description)
    else:
        path_in = pathlib.Path(path).resolve()
    if recursive:
        # perform recursive analysis
        # path_in is now a list
        print("Recursing into directory tree... ", end="", flush=True)
        path_list = recursive_search(path=path_in)
        if not path_list:
            msg = "Recursive search did not find any series data in " \
                  + "'{}'!".format(path_in)
            raise IOError(msg)
        print("Done.")
        # path_out is set to None when recursive search is used
        path_out = None
        for ii, pi in enumerate(path_list):
            # request necessary metadata for each measurement
            # before the actual analysis is done.
            print("Input {}/{}: {}".format(ii+1, len(path_list), pi))
            main(path=pi, profile=profile, req_meta=req_meta)
        path_in = path_list
    else:
        # verify data set
        try:
            qpformat.load_data(path_in)
        except qpformat.BadFileFormatError as e:
            msg = "For a recursive search, please pass the " \
                  + "command line parameter '-r'."
            e.args = ("; ".join(list(e.args) + [msg]),)
            raise
        # perform regular analysis
        path_out = path_in.with_name(path_in.name + OUTPUT_SUFFIX)
        path_out.mkdir(exist_ok=True)
        if path is None:
            # print directories only if taken from command line
            print("Input:  {}".format(path_in))
            print("Output: {}".format(path_out))
        if profile:
            # use a user-specified profile
            ppath = get_profile_path(profile)
            cfg_profile = config.ConfigFile(ppath)
            cfg_out = config.ConfigFile(path_out)
            cfg_out.update(cfg_profile)
        # get known meta data kwargs from dataset
        transfer_meta_data(path_in, path_out)
        # user input missing meta data keyword values
        for mm in sorted(req_meta):
            input_setting(path=path_out,
                          section="meta",
                          key=mm)
    return path_in, path_out


@functools.lru_cache(maxsize=32)  # cached to avoid multiple prints
def parse(description="DryMass analysis."):
    """Obtain the input data set path by parsing the command line"""
    print("DryMass version {}".format(version))
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('path', metavar='path', nargs='+', type=str,
                        help='Data path')
    parser.add_argument("-p", "--profile",
                        help="Use an existing profile for data analysis. "
                             + "The profile can be either a path to a "
                             + "'drymass.cfg' file or profile in the "
                             + "local library (see `dm_profile` command).",
                        default=None,
                        type=str)
    parser.add_argument("-r", "--recursive",
                        help="Recursively search for measurement data "
                             + "and run DryMass separately for each folder.",
                        default=False,
                        action='store_true')
    args = parser.parse_args()
    # Workaround: We use nargs='+' and join the input to support white
    # spaces in path names.
    jpath = " ".join(args.path)
    path_in = pathlib.Path(jpath).resolve()
    if args.recursive and not path_in.is_dir():
        msg = "Given path must be directory in recursive mode; " \
              + "got '{}'!".format(path_in)
        raise ValueError(msg)
    return path_in, args.profile, args.recursive


def recursive_search(path):
    """Perform recursive search for supported measurements"""
    path = pathlib.Path(path).resolve()
    path_in = []
    # Get all candidates
    cands1 = list(path.rglob("*")) + [path]
    # Exclude all directories with the suffix _dm that contain drymass.cfg
    cands2 = []
    for c1 in cands1:
        tocheck = list(c1.parents)
        if c1.is_dir():
            tocheck.append(c1)
        for pp in tocheck:
            if pp.name.endswith("_dm") and (pp / "drymass.cfg").exists():
                break
        else:
            cands2.append(c1)
    # Determine all directory-based measurements (SeriesFolder format)
    for c2 in cands2:
        if c2.is_dir():
            try:
                ds = qpformat.load_data(path=c2, fmt="SeriesFolder")
            except (NotImplementedError, qpformat.BadFileFormatError):
                pass
            else:
                if len(ds) > 1:
                    path_in.append(c2)
    if not path_in:
        # If we have found SeriesFolder measurements, then we probably do not
        # want to analyze any other files lying around, especially
        # because they could be related to background correction.
        # Try to open all the other files with qpformat
        for c2 in cands2:
            if c2.is_file():
                try:
                    ds = qpformat.load_data(path=c2)
                except qpformat.file_formats.UnknownFileFormatError:
                    pass
                else:
                    if ds.is_series:
                        # The same thing that applies to SeriesFolder also
                        # applies to SeriesData in general.
                        path_in.append(c2)
                        # Note that we completely ignore all SingleData
                        # file formats here.
    return sorted(path_in)


def transfer_meta_data(path_in, path_out):
    """Read input meta data and write it to the configuration file"""
    ds = qpformat.load_data(path=path_in)
    cfg = config.ConfigFile(path_out)
    sec = cfg["meta"]
    for key in sorted(META_MAPPER):
        dskey, mult = META_MAPPER[key]
        if (key not in sec or sec[key] is None) and dskey in ds.meta_data:
            cfg.set_value("meta", key, ds.meta_data[dskey] * mult)



import pathlib
import shutil
import tempfile

import numpy as np
import qpimage

from drymass.cli import config, dialog


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, method="edge", model="projection",
                    refraction_increment=.18, num=1, write_config=True):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp), which_data="phase,amplitude")
    path_in = tempfile.mktemp(suffix=".h5", prefix="drymass_test_cli_dialog")
    path_in = pathlib.Path(path_in)
    with qpimage.QPSeries(h5file=path_in, h5mode="w", identifier="tes") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    path_out = path_in.with_name(path_in.name + dialog.OUTPUT_SUFFIX)
    path_out.mkdir()
    if write_config:
        # add drymass configuration file
        cfg = config.ConfigFile(path_out)
        cfg.set_value(section="meta", key="pixel size um", value=pxsize*1e6)
        cfg.set_value(section="meta", key="wavelength nm",
                      value=wavelength*1e9)
        cfg.set_value(section="meta", key="medium index", value=medium_index)
        cfg.set_value(section="sphere", key="method", value=method)
        cfg.set_value(section="sphere", key="model", value=model)
        cfg.set_value(section="sphere", key="refraction increment",
                      value=refraction_increment)
    return qpi, path_in, path_out


def test_merge_config():
    # test data
    _, path_in, path_out = setup_test_data(medium_index=1.30,
                                           pxsize=1.12e-6,
                                           method="image",
                                           model="projection",
                                           refraction_increment=1.1)

    # profile
    _, path_profile = tempfile.mkstemp(
        prefix="drymass_test_config_", suffix=".cfg")
    cfg = config.ConfigFile(path=path_profile)
    cfg.path.write_text("[meta]\n"
                        + "medium index = 1.31\n"
                        + "pixel size um = None\n"
                        + "[sphere]\n"
                        + "method = image\n"
                        + "model = rytov\n")

    cfg2 = config.ConfigFile(path=path_out)
    assert cfg2["sphere"]["refraction increment"] == 1.1, "for test case"
    assert cfg2["meta"]["pixel size um"] == 1.12, "for test case"

    # apply profile (merge with original configuration)
    dialog.main(path=path_in, profile=path_profile)

    # Sanity checks in case DryMass defaults changed
    assert cfg["sphere"]["refraction increment"] != 1.1, "for test case"
    assert cfg["meta"]["pixel size um"] != 1.12, "for test case"

    # The following three are all valid by just copying cfg to path_out
    assert cfg2["meta"]["medium index"] == 1.31
    assert cfg2["sphere"]["method"] == "image"
    assert cfg2["sphere"]["model"] == "rytov"
    # This one is only valid when the configs are merged
    assert cfg2["sphere"]["refraction increment"] == 1.1
    # This one is only valid when Nones in profile do not override path_out
    assert cfg2["meta"]["pixel size um"] == 1.12

    try:
        cfg.path.unlink()
        cfg2.path.unlink()
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import copy

import matplotlib as mpl
import matplotlib.pylab as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

from ..import extractroi


CM_PHASE = "viridis"
CM_PHASE_ERROR = copy.copy(plt.get_cmap("seismic"))
CM_PHASE_ERROR.set_over("#37FF32")
CM_PHASE_ERROR.set_under("#05C500")
CM_INTENSITY = "gray"
CM_REFRACTIVE_INDEX = "gnuplot2"


try:
    mpl.rcParams['font.family'] = "sans-serif"
    mpl.rcParams["text.usetex"] = False
except TypeError:  # building the docs
    pass


def add_cbar(ax, mapper, fmt="%.2f", units="",
             loc="right", size="5%", labelloc=None,
             extend="neither"):
    """Add a colorbar to a plot"""
    if labelloc is None:
        labelloc = loc
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(loc, size=size, pad=0.05)
    if units:
        cax.set_title(units, ha="left", loc="center", fontsize=11)
    acbar = plt.colorbar(mapper, cax=cax, format=fmt, extend=extend)
    acbar.ax.yaxis.set_ticks_position(labelloc)
    acbar.ax.yaxis.set_label_position(labelloc)
    return acbar


def plot_image(data, ax=None, imtype="phase", cbar=True, px_um=None,
               ret_cbar=False, cbformat=None, **kwargs):
    """Plot an image

    Parameters
    ----------
    data: 2d np.ndarray
        Input image
    ax: matplotlib.Axes
        Axis to plot to
    imtype: str
        One of ["intensity", "phase", "phase error",
        "refractive index"].
    cbar: bool
        Whether to add a colorbar.
    px_um: float
        Pixel size [µm]
    ret_cbar: bool
        Whether to return the colorbar.
    kwargs: dict
        Keyword arguments to `plt.imshow`.

    Returns
    -------
    ax [, cbar]:
        Axis and colorbar.
    """
    if ax is None:
        ax = plt.subplot(111)

    cbkw = {}
    if cbformat is None:
        cbkw["fmt"] = "%.3f"
    else:
        cbkw["fmt"] = cbformat

    if imtype == "phase":
        cmap = CM_PHASE
        gridcolor = "w"
        if cbformat is None:
            cbkw["fmt"] = "%.1f"
        cbkw["units"] = "[rad]"
    elif imtype == "intensity":
        cmap = CM_INTENSITY
        gridcolor = "k"
        cbkw["units"] = "[a.u.]"
        # Make sure gray is at 1 in the colormap
        if "vmin" in kwargs and "vmax" in kwargs:
            vmin = kwargs["vmin"]
            vmax = kwargs["vmax"]
            if vmin < 1 and vmax > 1:
                diff = max(1 - vmin, vmax - 1)
                kwargs["vmin"] = 1 - diff
                kwargs["vmax"] = 1 + diff
    elif imtype == "refractive index":
        cmap = CM_REFRACTIVE_INDEX
        gridcolor = "w"
        cbkw["units"] = ""
    elif imtype == "phase error":
        cmap = CM_PHASE_ERROR
        gridcolor = "k"
        if "vmin" not in kwargs and "vmax" not in kwargs:
            vmax = np.max(np.abs(data))
            kwargs["vmax"] = vmax
            kwargs["vmin"] = -vmax
        cbkw["units"] = "[rad]"
        cbkw["extend"] = "both"
    else:
        raise ValueError("Unknown image type: {}".format(imtype))

    if px_um is None:
        shx, shy = np.array(data.shape)
        unit = "px"
    else:
        shx, shy = np.array(data.shape) * px_um
        unit = "µm"

    mapper = ax.imshow(data,
                       cmap=cmap,
                       extent=(0, shy, 0, shx),
                       interpolation="bilinear",
                       origin="lower",
                       **kwargs)

    ax.set_xlabel("x [{}]".format(unit))
    ax.set_ylabel("y [{}]".format(unit))
    ax.grid(color=gridcolor, lw="1", alpha=.1)

    retval = [ax]

    if cbar:
        acbar = add_cbar(ax=ax,
                         mapper=mapper,
                         **cbkw)

        if ret_cbar:
            retval.append(acbar)

    if len(retval) == 1:
        return retval[0]
    else:
        return retval


def plot_qpi_phase(qpi, rois=None, path=None, labels_excluded=[]):
    """Plot phase data"""
    fig = plt.figure(figsize=(6, 4))
    ax1 = plt.subplot(111)
    px_um = qpi["pixel size"] * 1e6
    plot_image(data=qpi.pha,
               ax=ax1,
               imtype="phase",
               cbar=True,
               px_um=px_um)
    if rois:
        for roi in rois:
            slx, sly = roi.roi_slice
            x0 = slx.start * px_um
            x1 = slx.stop * px_um
            y0 = sly.start * px_um
            y1 = sly.stop * px_um

            if extractroi.is_ignored_roi(roi=roi, ignore_data=labels_excluded):
                color = "r"
                ax1.text(y1, x1, "excluded",
                         horizontalalignment="right",
                         verticalalignment="top",
                         color=color)
            else:
                color = "w"
            box = mpl.patches.Rectangle(xy=(y0, x0),
                                        width=y1 - y0,
                                        height=x1 - x0,
                                        facecolor="none",
                                        edgecolor=color,
                                        )
            ax1.add_patch(box)
            ax1.text(y0, x0, roi.identifier,
                     horizontalalignment="left",
                     verticalalignment="bottom",
                     color=color)
    plt.tight_layout(rect=(0, 0, 1, .93), pad=.1)

    fig.text(x=.5, y=.99, s="sensor phase image",
             verticalalignment="top",
             horizontalalignment="center",
             fontsize=14)

    if path:
        fig.savefig(path)
        plt.close()
    else:
        return fig


def plot_qpi_sphere(qpi_real, qpi_sim, path=None, simtype="simulation"):
    """Plot QPI sphere analysis data"""
    fig = plt.figure(figsize=(9, 5))
    px_um = qpi_real["pixel size"] * 1e6
    radius_um = qpi_sim["sim radius"] * 1e6
    center = qpi_sim["sim center"]
    index = qpi_sim["sim index"]

    real_phase = qpi_real.pha
    kw_phase = {"px_um": px_um,
                "cbar": True,
                "imtype": "phase",
                "vmin": real_phase.min(),
                "vmax": real_phase.max(),
                }

    real_inten = qpi_real.amp**2
    kw_inten = {"px_um": px_um,
                "cbar": True,
                "imtype": "intensity",
                "vmin": real_inten.min(),
                "vmax": real_inten.max(),
                }
    # real phase
    ax1 = plt.subplot(231, title="data (phase)")
    plot_image(data=real_phase, ax=ax1, **kw_phase)

    # simulated phase
    ax2 = plt.subplot(232, title=simtype + " (phase)")
    plot_image(data=qpi_sim.pha, ax=ax2, **kw_phase)
    ax2.text(0.01, .99,
             "index: {:.5f}\n".format(index)
             + "radius: {:.3f}µm".format(radius_um),
             horizontalalignment="left",
             verticalalignment="top",
             color="w",
             transform=ax2.transAxes,
             )

    # phase residuals
    ax3 = plt.subplot(233, title="phase residuals")
    errmax = qpi_sim.pha.max() * .2
    plot_image(data=qpi_sim.pha - real_phase, ax=ax3,
               imtype="phase error", vmax=errmax, vmin=-errmax,
               px_um=px_um)

    # real intensity
    ax4 = plt.subplot(234, title="data (intensity)")
    plot_image(data=real_inten, ax=ax4, **kw_inten)

    # computed intensity
    ax5 = plt.subplot(235)
    if len(simtype) > 9:
        # sometimes the title is too long and is printed on top of the units
        kw5 = {"loc": "right",
               "ha": "right"}
    else:
        kw5 = {}
    ax5.set_title(simtype + " (intensity)", **kw5)

    plot_image(data=qpi_sim.amp**2, ax=ax5, **kw_inten)

    # plot detected radius
    for ax in [ax1, ax2, ax4, ax5]:
        circ = mpl.patches.Circle(xy=((center[1] + .5) * px_um,
                                      (center[0] + .5) * px_um),
                                  radius=radius_um,
                                  facecolor="none",
                                  edgecolor="w",
                                  ls=(0, (3, 8)),
                                  lw=.5,
                                  )
        ax.add_patch(circ)

    # line plot through center
    ax6 = plt.subplot(236, title="phase line plot")
    if int(center[0]) >= 0 and int(center[0]) < qpi_sim.shape[0]:
        x = np.arange(qpi_real.shape[1]) * px_um
        ax6.plot(x, qpi_sim.pha[int(center[0])], label=simtype)
        ax6.plot(x, qpi_real.pha[int(center[0])], label="data")
    ax6.set_xlabel("[µm]")
    ax6.legend(loc="center right")

    # remove unused labels
    for ax in [ax1, ax2, ax3]:
        ax.set_xlabel("")
    for ax in [ax2, ax3, ax5]:
        ax.set_ylabel("")

    plt.tight_layout(rect=(0, 0, 1, .93), pad=.1, h_pad=.6)

    # add identifier
    fig.text(x=.5, y=.99, s=qpi_sim["identifier"],
             verticalalignment="top",
             horizontalalignment="center",
             fontsize=14)

    if path:
        fig.savefig(path)
        plt.close()
    else:
        return fig



import pathlib

from . import definitions
from . import parse_funcs
from .._version import version

#: DryMass configuration file name
FILE_CONFIG = "drymass.cfg"


class ConfigFile(object):
    def __init__(self, path):
        """DryMass configuration file management

        Manage configuration file of an experimental data set
        with restrictions imposed by
        :data:`drymass.cli.definitions.config`.

        Parameters
        ----------
        path: str
            path to the configuration file or a folder containing the
            configuration file :data:`FILE_CONFIG`.
        """
        path = pathlib.Path(path).resolve()
        if path.is_dir():
            path = path / FILE_CONFIG
        if not path.exists():
            path.touch()
        self.path = path

    def __getitem__(self, section):
        """Get a configuration section

        Parameters
        ----------
        section: str
            the configuration section

        Returns
        -------
        sectiondict: dict
            the configuration section dictionary
        """
        datadict = self._parse()
        if section in datadict:
            return datadict[section]
        elif section in definitions.config:
            # return default values
            secd = {}
            for kk in definitions.config[section]:
                secd[kk] = definitions.config[section][kk][0]
            # write result
            datadict[section] = secd
            self._write(datadict)
            return secd
        else:
            raise ValueError("Unknown section title: {}".format(section))

    def __setitem__(self, section, sectiondict):
        """Replace a section in the configuration file

        Parameters
        ----------
        section: str
            the section name
        sectiondict: dict
            the configuration dictionary

        Notes
        -----
        The previous content of the configuration section
        in the configuration file is replaced.
        """
        datadict = self._parse()
        for key in sectiondict:
            self._check_value(section, key, sectiondict[key])
        datadict[section] = sectiondict
        self._write(datadict)

    def _check_value(self, section, key, value):
        """Check if a section/key/value pair is valid

        Raises `ValueError` if this is not the case.
        Returns `None`.
        """
        if section not in definitions.config:
            raise ValueError("Unknown section title: {}".format(section))
        if key not in definitions.config[section]:
            raise ValueError("Unknown key: {}: {}".format(section, key))
        # For versions > 0.8.1, unknown configuration keys are `None`.
        # Prior versions also used `np.nan`. Keep "nan" in the list
        # below for backwards compatibility.
        if value in [None, "nan", "none", "None", "()", "[]"]:
            if definitions.config[section][key][0] is not None:
                msg = "Unset value '{}' not allowed for [{}]: {}!".format(
                      value, section, key)
                raise ValueError(msg)
            ret_value = None
        else:
            type_func = definitions.config[section][key][1]
            try:
                type_func(value)
            except BaseException as e:
                msg = "Failed to parse: '[{}]: {}={}'".format(section, key,
                                                              value)
                e.args = ("{}; {}".format(msg, ", ".join(e.args)),)
                raise
            ret_value = type_func(value)
        return ret_value

    def _parse_compat(self, section, key, value):
        if section == "bg":
            # drymass < 0.1.3: API changed in qpimage 0.1.6
            if (key in ["amplitude profile", "phase profile"]
                    and value == "ramp"):
                value = "tilt"
        elif section == "roi":
            # drymass <= 0.1.5: keys were renamed to reflect pixel units
            if key in ["dist border", "exclude overlap", "pad border"]:
                key += " px"
        return key, value

    def _parse(self, autocomplete=True):
        """Return configuration dictionary

        Parameters
        ----------
        autocomplete: bool
            whether to fill in default configuration values when
            the corresponding keys are missing in a given section.
            Note that missing sections are not added. If missing
            keys are found, the original configuration file is
            overridden with the new data. Disabling autocompletion
            also prevents writing to the configuration file.

        Returns
        -------
        datadict: dict of dicts
            configuration dictionary

        Notes
        -----
        This function is private, because the autocomplete feature is
        actually a desired behavior to keep the configuration file
        human-readable. Normal users should not be able to use it,
        because the concept could be considered confusing.
        """
        with self.path.open() as fd:
            data = fd.readlines()
        outdict = {}
        for line in data:
            line = line.strip()
            if (line.startswith("#") or
                    len(line) == 0):
                pass
            elif line.startswith("["):
                sec = line.strip("[]")
                outdict[sec] = {}
            else:
                key, val = line.split("=")
                key = key.strip()
                val = val.strip()
                # backwards compatibility:
                key, val = self._parse_compat(sec, key, val)
                val = self._check_value(sec, key, val)
                outdict[sec][key] = val
        if autocomplete:
            # Insert default variables where missing
            must_write = False
            for sec in outdict:
                for key in definitions.config[sec]:
                    if key not in outdict[sec]:
                        outdict[sec][key] = definitions.config[sec][key][0]
                        must_write = True
            if must_write:
                # Update the configuration file
                self._write(outdict)
        return outdict

    def _write(self, datadict):
        """Write configuration dictionary

        Parameters
        ----------
        datadict: dict of dicts
            the full configuration

        Notes
        -----
        The configuration key values are converted to the correct
        dtype before writing using the definitions given in
        definitions.py.
        """
        keys = sorted(list(datadict.keys()))
        lines = ["# DryMass version {}".format(version),
                 "# Configuration file documented at: ",
                 "# https://drymass.readthedocs.io/en/stable/"
                 + "sec_gs_configuration_file.html",
                 ]
        for kk in keys:
            lines.append("")
            lines.append("[{}]".format(kk))
            subkeys = sorted(list(datadict[kk].keys()))
            for sk in subkeys:
                value = datadict[kk][sk]
                typefunc = definitions.config[kk][sk][1]
                if value is not None:
                    value = typefunc(value)
                    if typefunc in [parse_funcs.strlist,
                                    parse_funcs.strlist_fsort]:
                        # cosmetics for e.g. '[roi]: ignore data'
                        value = ", ".join(value)
                lines.append("{} = {}".format(sk, value))
        for ii in range(len(lines)):
            lines[ii] += "\n"
        with self.path.open("w") as fd:
            fd.writelines(lines)

    def remove_section(self, section):
        """Remove a section from the configuration file"""
        datadict = self._parse(autocomplete=False)
        datadict.pop(section)
        self._write(datadict)

    def set_value(self, section, key, value):
        """Set a configuration key value

        Parameters
        ----------
        section: str
            the configuration section
        key: str
            the configuration key in `section`
        value:
            the configuration key value

        Notes
        -----
        Valid section and key names are defined in definitions.py
        """
        # load, update, and save
        sec = self[section]
        sec[key] = value
        self[section] = sec

    def update(self, other):
        """Update the current configuration with data from another

        Parameters
        ----------
        other: ConfigFile
            the configuration file from which data is imported into
            the current configuration

        Notes
        -----
        None-valued keys are ignored.
        """
        other_dict = other._parse(autocomplete=False)
        for sec in other_dict:
            for key in other_dict[sec]:
                value = other_dict[sec][key]
                if value is not None:
                    self.set_value(section=sec,
                                   key=key,
                                   value=value)



import io
from os import fspath

import matplotlib.image as mpimg
import qpimage
from skimage.external import tifffile

from ..anasphere import analyze_sphere

from . import plot
from . import config
from . import dialog
from .extracting import cli_extract_roi


#: Matplotlib images of sphere analysis (TIFF)
FILE_SPHERE_ANALYSIS_IMAGE = "sphere_{}_{}_images.tif"


def cli_analyze_sphere(path=None, ret_data=False, profile=None):
    """Perform sphere analysis"""
    description = "Determine integral refractive index, radius, and " \
                  + "related parameters by inferring spherical symmetry " \
                  + "for each phase object found."
    path_in, path_out = dialog.main(path=path,
                                    req_meta=["medium index",
                                              "pixel size um",
                                              "wavelength nm"],
                                    description=description,
                                    profile=profile)
    if isinstance(path_in, list):
        # recursive analysis
        for ii, pi in enumerate(path_in):
            print("Analyzing dataset {}/{}.".format(ii+1, len(path_in)))
            cli_analyze_sphere(path=pi)
        # nothing else to do
        return
    cfg = config.ConfigFile(path_out)
    h5roi = cli_extract_roi(path=path_in, ret_data=True)
    print("Performing sphere analysis... ", end="", flush=True)

    # canny edge detection parameters
    edgekw = {
        "clip_rmin": cfg["sphere"]["edge clip radius min"],
        "clip_rmax": cfg["sphere"]["edge clip radius max"],
        "mult_coarse": cfg["sphere"]["edge coarse"],
        "mult_fine": cfg["sphere"]["edge fine"],
        "maxiter": cfg["sphere"]["edge iter"],
    }
    # image fitting parameters
    imagekw = {
        "crel": cfg["sphere"]["image fit range position"],
        "rrel": cfg["sphere"]["image fit range radius"],
        "nrel": cfg["sphere"]["image fit range refractive index"],
        "fix_pha_offset": cfg["sphere"]["image fix phase offset"],
        "max_iter": cfg["sphere"]["image iter"],
        "stop_dc": cfg["sphere"]["image stop delta position"],
        "stop_dr": cfg["sphere"]["image stop delta radius"],
        "stop_dn": cfg["sphere"]["image stop delta refractive index"],
        "verbose": cfg["sphere"]["image verbosity"],
    }
    h5sim, changed, reused = analyze_sphere(
        h5roi=h5roi,
        dir_out=path_out,
        r0=cfg["specimen"]["size um"] / 2 * 1e-6,
        method=cfg["sphere"]["method"],
        model=cfg["sphere"]["model"],
        alpha=cfg["sphere"]["refraction increment"],
        rad_fact=cfg["sphere"]["radial inclusion factor"],
        edgekw=edgekw,
        imagekw=imagekw,
        ret_changed=True,
        ret_reused=True,
        )

    if reused:
        print("Done (reused {} previous fits).".format(reused))
    else:
        print("Done.")

    tifout = path_out / FILE_SPHERE_ANALYSIS_IMAGE.format(
        cfg["sphere"]["method"],
        cfg["sphere"]["model"]
    )
    if (changed and cfg["output"]["sphere images"]) or not tifout.exists():
        print("Plotting sphere images... ", end="", flush=True)
        # plot h5series and rmgr with matplotlib
        with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps_roi, \
                qpimage.QPSeries(h5file=h5sim, h5mode="r") as qps_sim, \
                tifffile.TiffWriter(fspath(tifout), imagej=True) as tf:
            for qpi_real in qps_roi:
                qpi_sim = find_qpi_by_identifier(qps_sim,
                                                 qpi_real["identifier"])
                if qpi_sim is not None:
                    assert qpi_real["identifier"] in qpi_sim["identifier"]
                    imio = io.BytesIO()
                    plot.plot_qpi_sphere(qpi_real=qpi_real,
                                         qpi_sim=qpi_sim,
                                         path=imio,
                                         simtype=cfg["sphere"]["model"])
                    imio.seek(0)
                    imdat = (mpimg.imread(imio) * 255).astype("uint8")
                    tf.save(imdat, compress=9)
        print("Done")

    if ret_data:
        return h5sim


def find_qpi_by_identifier(qps, identifier):
    """Find the QPImage that has the same initial parts as identifier

    Parameters
    ----------
    qps: qpimage.QPSereis
        Series data
    identifier: str
        A string "hash:number" that identifies a QPImage

    Returns
    -------
    qpi: qpimage.QPImage or None
        The `QPImage` in ``qps`` whose first two identifier parts match
        ``identifier`` or `None`, if no matching `QPImage` is found.
    """
    numid = identifier.count(":") + 1
    for qpi in qps:
        parts = qpi["identifier"].split(":")
        if ":".join(parts[:numid]) == identifier:
            break
    else:
        qpi = None
    return qpi



import pathlib
import shutil
import tempfile

import numpy as np
import qpimage

from drymass.cli import config, dialog
import qpformat


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1, write_config=True, path_in=None):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    if path_in is None:
        path_in = tempfile.mktemp(suffix=".h5",
                                  prefix="drymass_test_cli_convert")
    path_in = pathlib.Path(path_in)
    with qpimage.QPSeries(h5file=path_in, h5mode="w", identifier="tes") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    path_out = path_in.with_name(path_in.name + dialog.OUTPUT_SUFFIX)
    path_out.mkdir()
    if write_config:
        # add drymass configuration file
        cfg = config.ConfigFile(path_out)
        cfg.set_value(section="meta", key="pixel size um", value=pxsize*1e6)
        cfg.set_value(section="meta", key="wavelength nm",
                      value=wavelength*1e9)
        cfg.set_value(section="meta", key="medium index", value=medium_index)
    return qpi, path_in, path_out


def setup_test_data_recursive(n=2, **kwargs):
    path = tempfile.mkdtemp(prefix="drymass_test_cli_recurse_")
    path = pathlib.Path(path)
    for ii in range(n):
        path_in = path / "sub_{}.h5".format(ii)
        setup_test_data(path_in=path_in, **kwargs)
    return path


def test_simple_qpseries():
    path = setup_test_data_recursive(n=2, num=3)

    ps = dialog.recursive_search(path)
    assert len(ps) == 2

    # make sure files are the right ones
    for ii in range(len(ps)):
        with qpimage.QPSeries(h5file=ps[ii], h5mode="r") as qps:
            assert len(qps) == 3

    shutil.rmtree(path, ignore_errors=True)


def test_qpseries_with_bad_data():
    path = setup_test_data_recursive(n=2, num=3)

    (path / "bad_data.h5").write_text("This file should be ignored.")

    ps = dialog.recursive_search(path)
    assert len(ps) == 2

    # make sure files are the right ones
    for ii in range(len(ps)):
        with qpimage.QPSeries(h5file=ps[ii], h5mode="r") as qps:
            assert len(qps) == 3

    shutil.rmtree(path, ignore_errors=True)


def test_complex_qpseries():
    path = tempfile.mkdtemp(prefix="drymass_test_cli_recurse2_")
    path = pathlib.Path(path)
    paths = [path / "1" / "2",
             path / "3",
             path / "4" / "5" / "6"
             ]
    for pp in paths:
        pp.mkdir(parents=True, exist_ok=True)
        setup_test_data(path_in=pp / "test.h5")

    ps = dialog.recursive_search(path)
    for ii, pi in enumerate(ps):
        assert pi == paths[ii] / "test.h5"

    shutil.rmtree(path, ignore_errors=True)


def test_complex_folder_with_unused_qpseries():
    path = tempfile.mkdtemp(prefix="drymass_test_cli_recurse2_")
    path = pathlib.Path(path)
    qpi, _, _ = setup_test_data(num=1, path_in=path / "test.h5")

    path = pathlib.Path(path)
    paths = [path / "1" / "2",
             path / "3",
             path / "4" / "5" / "6"
             ]
    for pp in paths:
        pp.mkdir(parents=True, exist_ok=True)
        for ii in range(4):
            qpi.copy(h5file=pp / "test_{}.h5".format(ii))

    ps = dialog.recursive_search(path)
    for ii, pi in enumerate(ps):
        assert pi == paths[ii]
    shutil.rmtree(path, ignore_errors=True)


def test_recursive_root_include1():
    qpi, path_in, path_out = setup_test_data(num=2)
    ps = dialog.recursive_search(path_in)
    assert len(ps) == 1
    assert ps[0] == path_in
    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


def test_recursive_root_include2():
    path = tempfile.mkdtemp(prefix="drymass_test_cli_recurse2_")
    path = pathlib.Path(path)
    # this file will be ignored
    qpi, ignored_path, _ = setup_test_data(num=1, path_in=path / "test.h5")

    for ii in range(4):
        qpi.copy(h5file=path / "test_{}.h5".format(ii))

    ps = dialog.recursive_search(path)
    assert len(ps) == 1
    assert ignored_path not in ps

    ds = qpformat.load_data(path=path)
    assert len(ds) == 4

    shutil.rmtree(path, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
